"""
Trajectopy - Trajectory Evaluation in Python

Gereon Tombrink, 2023
mail@gtombrink.de
"""
import logging
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors
from matplotlib.figure import Figure
from trajectopy.settings.plot_settings import PlotSettings
from trajectopy.trajectory import Sorting
import seaborn as sns
import trajectopy.util.datahandling as datahandling
from trajectopy.evaluation.trajectory_deviations import DeviationCollection, AbsoluteTrajectoryDeviations
from trajectopy.plotting.util import norm_hist, scatter_plotter, stair_hist, vertical_subplots

logger = logging.getLogger("root")

XLABEL_DICT = {Sorting.CHRONO: "time [s]", Sorting.SPATIAL: "trajectory length [m]"}


def plot_combined_devs(devs: AbsoluteTrajectoryDeviations, plot_settings: PlotSettings = PlotSettings()) -> Figure:
    if devs.rot_dev is None:
        comb_devs = [[devs.comb_pos_devs * plot_settings.unit_multiplier]]
        ylabels = [f"position deviations {plot_settings.unit_str}"]
    else:
        comb_devs = [[devs.comb_pos_devs * plot_settings.unit_multiplier, np.rad2deg(devs.comb_rot_devs)]]
        ylabels = [f"position deviations {plot_settings.unit_str}", "minimum rotation angle [°]"]

    return vertical_subplots(
        x_list=[devs.function_of] * len(comb_devs),
        y_list=comb_devs,
        xlabel=XLABEL_DICT[devs.sorting],
        ylabels=ylabels,
        plot_settings=plot_settings,
    )


def plot_raw_rotation_devs(devs: AbsoluteTrajectoryDeviations, plot_settings: PlotSettings = PlotSettings()) -> Figure:
    rpy_dev = devs.rpy_dev
    vert_sp_data_rot = [
        [
            np.rad2deg(rpy_dev[:, 0]),
            np.rad2deg(rpy_dev[:, 1]),
            np.rad2deg(rpy_dev[:, 2]),
        ]
    ]
    ylabels = ["roll [°]", "pitch [°]", "yaw [°]"]

    return vertical_subplots(
        x_list=[devs.function_of] * len(vert_sp_data_rot),
        y_list=vert_sp_data_rot,
        xlabel=XLABEL_DICT[devs.sorting],
        ylabels=ylabels,
        plot_settings=plot_settings,
    )


def plot_raw_position_devs(devs: AbsoluteTrajectoryDeviations, plot_settings: PlotSettings = PlotSettings()) -> Figure:
    deviations_xa = devs.directed_pos_dev[:, 0] if plot_settings.show_directed_devs else devs.pos_dev[:, 0]
    deviations_yh = devs.directed_pos_dev[:, 1] if plot_settings.show_directed_devs else devs.pos_dev[:, 1]
    deviations_zv = devs.directed_pos_dev[:, 2] if plot_settings.show_directed_devs else devs.pos_dev[:, 2]

    if plot_settings.show_directed_devs:
        ylabels = [
            f"along {plot_settings.unit_str}",
            f"horiz. cross {plot_settings.unit_str}",
            f"vert. cross {plot_settings.unit_str}",
        ]
    else:
        ylabels = [f"x {plot_settings.unit_str}", f"y {plot_settings.unit_str}", f"z {plot_settings.unit_str}"]

    vert_sp_data_pos = [
        [
            deviations_xa * plot_settings.unit_multiplier,
            deviations_yh * plot_settings.unit_multiplier,
            deviations_zv * plot_settings.unit_multiplier,
        ]
    ]

    return vertical_subplots(
        x_list=[devs.function_of] * len(vert_sp_data_pos),
        y_list=vert_sp_data_pos,
        xlabel=XLABEL_DICT[devs.sorting],
        ylabels=ylabels,
        plot_settings=plot_settings,
    )


def plot_dof_dev(*, devs: AbsoluteTrajectoryDeviations, plot_settings: PlotSettings = PlotSettings()) -> Figure:
    fig = plt.figure(figsize=(12, 4))

    plot_position_dof(devs, plot_settings)

    if devs.rot_dev:
        plot_rot_dof(devs, plot_settings)

    return fig


def plot_position_dof(devs: AbsoluteTrajectoryDeviations, plot_settings: PlotSettings = PlotSettings()):
    deviations_xa = devs.directed_pos_dev[:, 0] if plot_settings.show_directed_devs else devs.pos_dev[:, 0]
    deviations_yh = devs.directed_pos_dev[:, 1] if plot_settings.show_directed_devs else devs.pos_dev[:, 1]
    deviations_zv = devs.directed_pos_dev[:, 2] if plot_settings.show_directed_devs else devs.pos_dev[:, 2]

    rows = 1 if devs.rot_dev is None else 2
    xyz_dev = [devs.pos.xyz]
    c_labels = [plot_settings.unit_str]

    plt.subplot(rows, 3, 1)
    scatter_plotter(
        xyz=xyz_dev,
        data=[deviations_xa * plot_settings.unit_multiplier],
        c_labels=c_labels,
        titles=["along-track" if plot_settings.show_directed_devs else "x"],
        figure=False,
        plot_settings=plot_settings,
    )

    plt.subplot(rows, 3, 2)
    scatter_plotter(
        xyz=xyz_dev,
        data=[deviations_yh * plot_settings.unit_multiplier],
        c_labels=c_labels,
        titles=["horizontal cross-track" if plot_settings.show_directed_devs else "y"],
        figure=False,
        plot_settings=plot_settings,
    )

    plt.subplot(rows, 3, 3)
    scatter_plotter(
        xyz=xyz_dev,
        data=[deviations_zv * plot_settings.unit_multiplier],
        c_labels=c_labels,
        titles=["vertical cross-track" if plot_settings.show_directed_devs else "z"],
        figure=False,
        plot_settings=plot_settings,
    )


def plot_rot_dof(devs: AbsoluteTrajectoryDeviations, plot_settings: PlotSettings = PlotSettings()) -> None:
    xyz_dev = [devs.pos.xyz]
    rpy_dev = devs.rpy_dev
    plt.subplot(2, 3, 4)
    scatter_plotter(
        xyz=xyz_dev,
        data=[np.rad2deg(rpy_dev[:, 0])],
        c_labels=["[°]"],
        titles=["roll"],
        figure=False,
        plot_settings=plot_settings,
    )

    plt.subplot(2, 3, 5)
    scatter_plotter(
        xyz=xyz_dev,
        data=[np.rad2deg(rpy_dev[:, 1])],
        c_labels=["[°]"],
        titles=["pitch"],
        figure=False,
        plot_settings=plot_settings,
    )

    plt.subplot(2, 3, 6)
    scatter_plotter(
        xyz=xyz_dev,
        data=[np.rad2deg(rpy_dev[:, 2])],
        c_labels=["[°]"],
        titles=["yaw"],
        figure=False,
        plot_settings=plot_settings,
    )


def compact_deviations(
    devs: AbsoluteTrajectoryDeviations, plot_settings: PlotSettings = PlotSettings()
) -> Figure | None:
    # positions
    dev_pos_list = [devs.pos.xyz]
    comb_pos_rms = datahandling.moving_width(
        t=devs.function_of,
        data=devs.comb_pos_devs,
        width=plot_settings.rms_window_width,
        function=datahandling.rms,
    )

    if devs.rot_dev:
        comb_rot_rms = datahandling.moving_width(
            t=devs.function_of,
            data=devs.comb_rot_devs,
            width=plot_settings.rms_window_width,
            function=datahandling.rms,
        )
        data_rms_plot = [comb_pos_rms * plot_settings.unit_multiplier, np.rad2deg(comb_rot_rms)]
        titles_rms = ["position rms", "rotation rms"]
        c_labels_dev = [plot_settings.unit_str, "[°]"]

    else:
        data_rms_plot = [comb_pos_rms * plot_settings.unit_multiplier]
        titles_rms = ["position rms"]
        c_labels_dev = [plot_settings.unit_str]

    return scatter_plotter(
        xyz=dev_pos_list * len(data_rms_plot),
        data=data_rms_plot,
        c_labels=c_labels_dev,
        titles=titles_rms,
        plot_settings=plot_settings,
    )


def plot_compact_hist(devs: AbsoluteTrajectoryDeviations, plot_settings: PlotSettings = PlotSettings()) -> Figure:
    """
    Plot compact histograms of cross-track and rpy deviations
    """
    fig = plt.figure()
    pos_ax = plt.subplot(2, 1, 1)
    plot_position_hist(devs, plot_settings)
    pos_ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))

    if devs.rot_dev is not None:
        rot_ax = plt.subplot(2, 1, 2)
        plot_rotation_hist(devs, plot_settings)
        rot_ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))

    return fig


def plot_rotation_hist(devs: AbsoluteTrajectoryDeviations, plot_settings: PlotSettings = PlotSettings()) -> None:
    roll = np.rad2deg(devs.rpy_dev[:, 0])
    pitch = np.rad2deg(devs.rpy_dev[:, 1])
    yaw = np.rad2deg(devs.rpy_dev[:, 2])

    plt.xlabel("[°]")
    plt.ylabel("counts")
    if plot_settings.hist_as_stairs:
        stair_hist(l=yaw)
        stair_hist(l=pitch)
        stair_hist(l=roll)
    else:
        norm_hist(l=yaw, alpha=0.6, norm=False)
        norm_hist(l=pitch, alpha=0.6, norm=False)
        norm_hist(l=roll, alpha=0.4, norm=False)
    plt.legend(["yaw", "pitch", "roll"])


def plot_position_hist(devs: AbsoluteTrajectoryDeviations, plot_settings: PlotSettings = PlotSettings()):
    deviations_xa = devs.directed_pos_dev[:, 0] if plot_settings.show_directed_devs else devs.pos_dev[:, 0]
    deviations_yh = devs.directed_pos_dev[:, 1] if plot_settings.show_directed_devs else devs.pos_dev[:, 1]
    deviations_zv = devs.directed_pos_dev[:, 2] if plot_settings.show_directed_devs else devs.pos_dev[:, 2]
    labels = ["vertical", "horizontal", "along"] if plot_settings.show_directed_devs else ["x", "y", "z"]

    plt.xlabel(plot_settings.unit_str)
    plt.ylabel("counts")
    if plot_settings.hist_as_stairs:
        stair_hist(l=deviations_zv, mm=plot_settings.unit_is_mm)
        stair_hist(l=deviations_yh, mm=plot_settings.unit_is_mm)
        stair_hist(l=deviations_xa, mm=plot_settings.unit_is_mm)
    else:
        norm_hist(l=deviations_zv, mm=plot_settings.unit_is_mm, norm=False)
        norm_hist(l=deviations_yh, mm=plot_settings.unit_is_mm, norm=False)
        norm_hist(l=deviations_xa, mm=plot_settings.unit_is_mm, norm=False)

    plt.legend(labels)


def plot_edf(
    deviation: AbsoluteTrajectoryDeviations | list[AbsoluteTrajectoryDeviations],
    plot_settings: PlotSettings = PlotSettings(),
) -> Figure:
    deviation_list = deviation if isinstance(deviation, list) else [deviation]

    fig = plt.figure()
    plot_position_edf(deviation_list, plot_settings)
    plot_rotation_edf(deviation_list, plot_settings)

    fig.legend([dev.name for dev in deviation_list], ncol=3, loc="upper center")
    return fig


def plot_position_edf(
    deviation_list: list[AbsoluteTrajectoryDeviations], plot_settings: PlotSettings = PlotSettings()
) -> None:
    ax_pos = plt.subplot(2, 1, 1)
    ax_pos.set_xlabel(f"Deviation {plot_settings.unit_str}")
    ax_pos.set_ylabel("Cummulative Probability")

    for dev in deviation_list:
        sorted_comb_pos_dev = np.sort(dev.comb_pos_devs)
        pos_norm_cdf = np.arange(len(sorted_comb_pos_dev)) / float(len(sorted_comb_pos_dev))
        ax_pos.plot(sorted_comb_pos_dev * plot_settings.unit_multiplier, pos_norm_cdf)


def plot_rotation_edf(
    deviation_list: list[AbsoluteTrajectoryDeviations], plot_settings: PlotSettings = PlotSettings()
) -> None:
    ax_rot = plt.subplot(2, 1, 2)

    ax_rot.set_xlabel("Deviation [°]")
    ax_rot.set_ylabel("Cummulative Probability")

    for dev in deviation_list:
        if dev.rot_dev is None:
            continue
        sorted_comb_rot_dev = np.sort(np.rad2deg(dev.comb_rot_devs))
        rot_norm_cdf = np.arange(len(sorted_comb_rot_dev)) / float(len(sorted_comb_rot_dev))
        ax_rot.plot(sorted_comb_rot_dev, rot_norm_cdf)


def plot_bars(
    deviation_list: list[AbsoluteTrajectoryDeviations],
    plot_settings: PlotSettings = PlotSettings(),
    mode: str = "positions",
) -> Figure:
    fig, ax = plt.subplots()
    bar_width = 0.9 / len(deviation_list)
    characteristics = ["Min", "Max", "Mean", "Median", "RMS", "STD"]
    unit = plot_settings.unit_str if mode == "positions" else "[°]"
    spacings = np.linspace(
        -bar_width * (len(deviation_list) - 1) / 2, bar_width * (len(deviation_list) - 1) / 2, len(deviation_list)
    )
    x_positions = np.arange(len(characteristics))
    for deviation, spacing in zip(deviation_list, spacings):
        if mode == "rotations" and deviation.rot_dev is None:
            continue

        if mode == "positions":
            data = [
                deviation.min_pos * plot_settings.unit_multiplier,
                deviation.max_pos * plot_settings.unit_multiplier,
                deviation.mean_pos * plot_settings.unit_multiplier,
                deviation.median_pos * plot_settings.unit_multiplier,
                deviation.rms_pos * plot_settings.unit_multiplier,
                deviation.std_pos * plot_settings.unit_multiplier,
            ]
        elif mode == "rotations":
            data = [
                np.rad2deg(deviation.min_rot),
                np.rad2deg(deviation.max_rot),
                np.rad2deg(deviation.mean_rot),
                np.rad2deg(deviation.median_rot),
                np.rad2deg(deviation.rms_rot),
                np.rad2deg(deviation.std_rot),
            ]
        else:
            raise ValueError("Mode must be either 'positions' or 'rotations'")
        ax.bar(x_positions + spacing, data, width=bar_width, label=deviation.name)

    ax.set_xlabel("Characteristic")
    ax.set_ylabel(f"Value {unit}")
    ax.set_xticks(x_positions)
    ax.set_xticklabels(characteristics)
    ax.legend()

    return fig


def rms_heatmap(
    deviation_collection: DeviationCollection, plot_settings: PlotSettings = PlotSettings()
) -> Tuple[Figure, Figure | None]:
    pos_rms_fig = plot_position_rms_heatmap(deviation_collection, plot_settings)
    rpy_rms_fig = plot_rotation_rms_heatmap(deviation_collection, plot_settings)

    return pos_rms_fig, rpy_rms_fig


def plot_rotation_rms_heatmap(
    deviation_collection: DeviationCollection, plot_settings: PlotSettings = PlotSettings()
) -> Figure | None:
    if not deviation_collection.rpy_rms:
        return None

    rpy_rms_fig, ax = plt.subplots()
    sns.heatmap(
        np.rad2deg(np.array(deviation_collection.rpy_rms)),
        ax=ax,
        annot=True,
        cmap="YlOrRd",
        fmt=".3f",
        linewidth=plot_settings.heatmap_spacing,
        cbar_kws={"label": "[°]"},
    )
    rpy_rms_fig.subplots_adjust(left=0.3)
    ax.set_xticklabels(["roll rms", "pitch rms", "yaw rms"], rotation=45)
    ax.set_yticklabels(deviation_collection.rpy_labels, rotation=0)

    return rpy_rms_fig


def plot_position_rms_heatmap(
    deviation_collection: DeviationCollection, plot_settings: PlotSettings = PlotSettings()
) -> Figure:
    pos_rms = (
        deviation_collection.directed_pos_rms if plot_settings.show_directed_devs else deviation_collection.pos_rms
    )

    pos_labels = ["along rms", "cross rms", "vertical rms"] if plot_settings.show_directed_devs else ["x", "y", "z"]
    pos_rms_fig, ax = plt.subplots()
    sns.heatmap(
        np.array(pos_rms) * plot_settings.unit_multiplier,
        ax=ax,
        annot=True,
        cmap="YlOrRd",
        fmt=".3f",
        linewidth=plot_settings.heatmap_spacing,
        cbar_kws={"label": plot_settings.unit_str},
    )
    pos_rms_fig.subplots_adjust(left=0.3)
    ax.set_xticklabels(pos_labels, rotation=45 if plot_settings.show_directed_devs else 0)
    ax.set_yticklabels(deviation_collection.names, rotation=0)

    return pos_rms_fig


def bias_heatmap(
    deviation_collection: DeviationCollection, plot_settings: PlotSettings = PlotSettings()
) -> Tuple[Figure, Figure | None]:
    pos_bias_fig = plot_position_bias_heatmap(deviation_collection, plot_settings)

    rpy_bias_fig = plot_rotation_bias_heatmap(deviation_collection, plot_settings)

    return pos_bias_fig, rpy_bias_fig


def plot_rotation_bias_heatmap(
    deviation_collection: DeviationCollection, plot_settings: PlotSettings = PlotSettings()
) -> Figure | None:
    if not deviation_collection.rpy_bias:
        return None

    # roll pitch yaw bias
    rpy_bias_fig, ax = plt.subplots()
    sns.heatmap(
        np.rad2deg(np.array(deviation_collection.rpy_bias)),
        ax=ax,
        annot=True,
        cmap="bwr",
        fmt=".3f",
        norm=colors.CenteredNorm(),
        linewidth=plot_settings.heatmap_spacing,
        cbar_kws={"label": "[°]"},
    )
    rpy_bias_fig.subplots_adjust(left=0.3)
    ax.set_xticklabels(["bias roll", "bias pitch", "bias yaw"], rotation=45)
    ax.set_yticklabels(deviation_collection.rpy_labels, rotation=0)

    return rpy_bias_fig


def plot_position_bias_heatmap(
    deviation_collection: DeviationCollection, plot_settings: PlotSettings = PlotSettings()
) -> Figure:
    pos_bias = (
        deviation_collection.directed_pos_bias if plot_settings.show_directed_devs else deviation_collection.pos_bias
    )

    pos_labels = ["along bias", "cross bias", "vertical bias"] if plot_settings.show_directed_devs else ["x", "y", "z"]
    pos_bias_fig, ax = plt.subplots()
    sns.heatmap(
        np.array(pos_bias) * plot_settings.unit_multiplier,
        ax=ax,
        annot=True,
        cmap="bwr",
        fmt=".3f",
        norm=colors.CenteredNorm(),
        linewidth=plot_settings.heatmap_spacing,
        cbar_kws={"label": plot_settings.unit_str},
    )
    pos_bias_fig.subplots_adjust(left=0.3)
    ax.set_xticklabels(pos_labels, rotation=45 if plot_settings.show_directed_devs else 0)
    ax.set_yticklabels(deviation_collection.names, rotation=0)
    return pos_bias_fig


def plot_multiple_comb_deviations(
    deviation: AbsoluteTrajectoryDeviations | list[AbsoluteTrajectoryDeviations],
    plot_settings: PlotSettings = PlotSettings(),
) -> Figure:
    deviation_list = deviation if isinstance(deviation, list) else [deviation]

    if all(dev.state.sorting_known for dev in deviation_list):
        x_label = "Trajectory length [m]"
        target_sorting = Sorting.SPATIAL
    else:
        target_sorting = Sorting.CHRONO
        x_label = "Time [s]"

    fig = plt.figure()
    ax_pos = plt.subplot(2, 1, 1)
    ax_pos.set_xlabel(x_label)
    ax_pos.set_ylabel(f"Deviation {plot_settings.unit_str}")

    if any(dev.rot_dev for dev in deviation_list):
        ax_rot = plt.subplot(2, 1, 2)
        ax_rot.set_xlabel(x_label)
        ax_rot.set_ylabel("Deviation [°]")
    else:
        ax_rot = None

    for dev in deviation_list:
        dev_sorted: AbsoluteTrajectoryDeviations = dev.set_sorting(target_sorting, inplace=False)
        arc_length_sorting = np.argsort(dev_sorted.function_of)
        ax_pos.plot(
            dev_sorted.function_of[arc_length_sorting],
            dev_sorted.comb_pos_devs[arc_length_sorting] * plot_settings.unit_multiplier,
        )
        if ax_rot is not None:
            ax_rot.plot(dev_sorted.function_of[arc_length_sorting], dev_sorted.comb_rot_devs[arc_length_sorting])

    fig.legend([dev.name for dev in deviation_list], ncol=3, loc="upper center")
    return fig


def plot_multiple_deviations(
    deviation_collection: DeviationCollection, plot_settings: PlotSettings = PlotSettings()
) -> Tuple[Figure, Figure | None]:
    if plot_settings.show_directed_devs:
        pos_labels = [
            f"along {plot_settings.unit_str}",
            f"horiz. cross {plot_settings.unit_str}",
            f"vert. cross {plot_settings.unit_str}",
        ]
    else:
        pos_labels = [f"x {plot_settings.unit_str}", f"y {plot_settings.unit_str}", f"z {plot_settings.unit_str}"]

    rpy_labels = ["roll [°]", "pitch [°]", "yaw [°]"]

    pos_dev = (
        deviation_collection.directed_pos_dev if plot_settings.show_directed_devs else deviation_collection.pos_dev
    )
    xyz_dev_fig = _plot_components(
        x_list=deviation_collection.lengths,
        y_list=pos_dev,
        ylabels=pos_labels,
        legend_list=deviation_collection.names,
        sharey=True,
        mm=plot_settings.unit_is_mm,
        plot_settings=plot_settings,
    )

    # rpy
    if deviation_collection.rpy_dev and deviation_collection.lengths:
        rpy_dev_fig = _plot_components(
            x_list=deviation_collection.lengths,
            y_list=deviation_collection.rpy_dev,
            ylabels=rpy_labels,
            legend_list=deviation_collection.rpy_labels,
            sharey=True,
            deg=True,
            plot_settings=plot_settings,
        )
    else:
        rpy_dev_fig = None

    return xyz_dev_fig, rpy_dev_fig


def _plot_components(
    *,
    x_list: list[np.ndarray],
    y_list: list[np.ndarray],
    ylabels: list,
    legend_list: list,
    sharey: bool = False,
    mm: bool = False,
    deg: bool = False,
    plot_settings: PlotSettings = PlotSettings(),
) -> Figure:
    """
    Plots data for each lap for each component in x
    """
    fig, axes = plt.subplots(nrows=y_list[0].shape[1], ncols=1, sharex=True, sharey=sharey)
    for j, (ax, yl) in enumerate(zip(axes, ylabels)):
        if j == y_list[0].shape[1] - 1:
            ax.set_xlabel("trajectory length [m]")
        ax.set_ylabel(yl)

        x_max = 0
        for x_l, l_l in zip(y_list, x_list):
            idx = np.argsort(l_l)
            if mm:
                data = x_l[idx, j] * 1000
            elif deg:
                data = np.rad2deg(np.unwrap(x_l[idx, j]))
            else:
                data = x_l[idx, j]

            ax.plot(l_l[idx], data)
            if max(l_l[idx]) > x_max:
                x_max = max(l_l[idx])
        ax.set_xlim([0, x_max])

    fig.legend(legend_list, ncol=3, loc="upper center")
    fig.subplots_adjust(right=0.85)
    return fig
